/**************************************

                CODE.C -- Routines to generate mach.
                language code.
***************************************/



#include "sys.h"
#include <dos.h>


/*
        opcodes for 808X instuctions
*/

#define PUSH_AX         0x50
#define MOV_AX          0xb8
#define INC_SP          0x44
#define FAR_CALL        0x9a
#define RET             0xcb            /* return from far call */
#define JMPS            0xeb            /* jump - 8 bit displacement */
#define JMPL            0xe9            /* jump - 16 bit displacement */
#define MOV_BX          0xbb
#define DEC_BX          0x4b
#define PUSH_INDBX1     0xff
#define PUSH_INDBX2     0x37
#define ADD_SP1         0x81
#define ADD_SP2         0xC4


#define DEBUG           FALSE
#define MAX_INSTRS      1000            /* max. # instructions compiled into
                                           one buffer                   */

char instrs[MAX_INSTRS];
char *pres_instr;
unsigned int code_seg, num_instr;

void init_code()
{
        void segread();

        struct SREG segs;

        segread(&segs);                 /* get value of code segment reg */
        code_seg = segs.cs;

        pres_instr = instrs;            /* code will be generated starting here */
        num_instr = 0;                  /* no instructions generated yet */
}



/*
        Emit one byte of code to the next available position.
*/
void codeb(op_code)
unsigned int op_code;
{
        *pres_instr = (char)(op_code);
        pres_instr++;
        num_instr++;
}


/*
        Emit two bytes (one word) of code to the next available position.
*/
void codew(word)
unsigned int word;
{
        codeb(word & 0xff);
        codeb(word >> 8);
}

void code_ret()
{
        codeb(RET);
#if DEBUG
        printf("        RET\n");
#endif
}

/*
        Code an assignment
*/
void code_assig(ptr)
DATA_TYPE *ptr;
{
        extern void assign();

        codeb(MOV_BX);          /*      MOV     BX, OFFSET PTR */
        codew(ptr);


        codeb(FAR_CALL);        /*      CALL    CODE_SEG:ASSIGN */
        codew(assign);
        codew(code_seg);

                        /* clean up the stack */

        codeb(ADD_SP1);         /*      ADD     SP, 8           */
        codeb(ADD_SP2);
        codew(sizeof(DATA_TYPE));
#if DEBUG
        printf("        mov     bx,%u\n", ptr);
        printf("        call    _assig\n");
        printf("        add     sp, %d\n", sizeof(DATA_TYPE));
#endif
}


/*
        code to get a variable's value (a cell's value) onto the stack
*/
void code_indirect(addr)
unsigned int addr;
{
        codeb(MOV_BX);
        codew(addr + 6);

        codeb(PUSH_INDBX1);     /* push first word of double */
        codeb(PUSH_INDBX2);

        codeb(DEC_BX);          /* push next word */
        codeb(DEC_BX);
        codeb(PUSH_INDBX1);
        codeb(PUSH_INDBX2);

        codeb(DEC_BX);          /* push next word */
        codeb(DEC_BX);
        codeb(PUSH_INDBX1);
        codeb(PUSH_INDBX2);

        codeb(DEC_BX);          /* push next word */
        codeb(DEC_BX);
        codeb(PUSH_INDBX1);
        codeb(PUSH_INDBX2);
}


void code_op(op)
token_type op;
{
        extern void mud(), dvd(), add(), sub();
        extern void clean(), neg();

        codeb(FAR_CALL);

        switch (op)
        {
                case '*':
#if DEBUG
                        printf("        call    _mud\n");
#endif
                        codew(mud);
                        break;

                case '/':
#if DEBUG
                        printf("        call    _dvd\n");
#endif
                        codew(dvd);
                        break;

                case '+':
#if DEBUG
                        printf("        call    _add\n");
#endif
                        codew(add);
                        break;

                case '-':
#if DEBUG
                        printf("        call    _sub\n");
#endif
                        codew(sub);
                        break;

                case UN_MINUS:
#if DEBUG
                        printf("        call    _neg\n");
#endif
                        codew(neg);
                        break;
        }
        codew(code_seg);
}

void code_jmp(loc)
unsigned int loc;
{
        unsigned int disp;

        disp = loc - ((unsigned int)(pres_instr) + 2);

                                        /* will a short jump do the job? */
/*        if((disp >> 8) != 0xff && (disp >> 8) != 0)

        {
*/
                disp--;
                codeb(JMPL);
                codew(disp);
#if DEBUG
                printf("        jmp    $%x  ; (jmp long to %x)\n",
                                              disp, loc);
#endif
/*        }
        else
        {
                codeb(JMPS);
                codeb(disp & 0xff);
#if DEBUG
                printf("        jmp    $%x  ; (jmp short to %x)\n",
                                             disp & 0xff, loc);
#endif
        }
*/
}


/*
        Code to push an immed. value onto the stack
*/
void code_push(m3, m2, m1, e)
unsigned int e, m1, m2, m3;
{
#if DEBUG
        printf("        mov     ax,%u\n", e);
        printf("        push    ax\n");
#endif
        codeb(MOV_AX);
        codew(e);
        codeb(PUSH_AX);

        if(m1 != e)
        {
#if DEBUG
                printf("        mov     ax,%u\n", m1);
#endif
                codeb(MOV_AX);
                codew(m1);
        }
#if DEBUG
        printf("        push    ax\n");
#endif
        codeb(PUSH_AX);

        if(m2 != m1)
        {
#if DEBUG
                printf("        mov     ax,%u\n", m2);
#endif
                codeb(MOV_AX);
                codew(m2);
        }
#if DEBUG
        printf("        push    ax\n");
#endif
        codeb(PUSH_AX);

        if(m3 != m2)
        {
#if DEBUG
        printf("        mov     ax,%u\n", m3);
#endif
                codeb(MOV_AX);
                codew(m3);
        }
#if DEBUG
        printf("        push    ax\n");
#endif
        codeb(PUSH_AX);
}

#define SIZE_LONG_JUMP 3        /* bytes to encode a far jump */

/*
        New code has been generated.  Copy it into a chunk of memory
        and leave a little extra room for a jump instruction.  The jump
        instruction will be patched in later to fit this new code into
        the code already extant.
*/
unsigned int new_code(ptr)
char **ptr;
{
        char *zalloc();

        char *source, *dest, *f;
        static unsigned int i;

        f = *ptr;
        *ptr = zalloc(num_instr + SIZE_LONG_JUMP);
        dest = *ptr;
        source = instrs;
        for(i = 0; i < num_instr; i++, dest++, source++)*dest = *source;
        for(i=0; i< SIZE_LONG_JUMP; i++)
        {
                *dest = 0;
                dest++;
        }
        if(f)free(f);
        return num_instr - 1;
}


/*
        Patch a jump instruction so that the program will continue execution
        at the correct place after the present chunk of code
*/
void new_jump(code, pos, loc)
char *code, *loc;
unsigned int pos;
{
        unsigned int disp;
        char *pres;

        pres = code + pos;
        disp = loc - ((unsigned int)(pres) + 3);

        *pres = JMPL;
        pres++;
        *pres = disp & 0xff;
        pres++;
        *pres = disp >> 8;
}
